@@ -19,6 +19,8 @@
 #include <linux/fs.h>
 #include <linux/major.h>
 
+#define VERSION "v3.6"
+
 /*
  * For kernels prior to 1.3.61, HDIO_SET_32BIT is HDIO_SET_CHIPSET
  */
@@ -32,14 +34,14 @@
 #define TIMING_MB		64
 #define TIMING_BUF_MB		1
 #define TIMING_BUF_BYTES	(TIMING_BUF_MB * 1024 * 1024)
-#define TIMING_BUF_COUNT	(TIMING_MB / TIMING_BUF_MB)
+#define TIMING_BUF_COUNT	(timing_MB / TIMING_BUF_MB)
 #define BUFCACHE_FACTOR		2
 
 char *progname;
 static struct hd_driveid id;
 static int verbose = 0, get_identity = 0, get_geom = 0, noisy = 1, quiet = 0;
-static int flagcount = 0, do_flush = 0, is_scsi_hd = 0;;
-static int do_ctimings, do_timings = 0, old_identity = 0;
+static flagcount = 0, do_flush = 0, is_scsi_hd = 0, is_xt_hd = 0;
+static int do_ctimings, do_timings = 0, old_identity = 0, timing_MB = TIMING_MB;
 
 static unsigned long set_readahead= 0, get_readahead= 0, readahead= 0;     
 static unsigned long set_readonly = 0, get_readonly = 0, readonly = 0;     
@@ -68,6 +70,7 @@ static unsigned long set_lookahead= 0, get_lookahead= 0, lookahead= 0;
 static unsigned long set_prefetch = 0, get_prefetch = 0, prefetch = 0;
 static unsigned long set_defects  = 0, defects = 0;
 static unsigned long set_wcache   = 0, get_wcache   = 0, wcache   = 0;
+static unsigned long set_doorlock = 0, get_doorlock = 0, doorlock = 0;
 static unsigned long set_seagate  = 0, get_seagate  = 0;
 static unsigned long set_standbynow = 0, get_standbynow = 0;
 static unsigned long set_sleepnow   = 0, get_sleepnow   = 0;
@@ -120,37 +123,18 @@ static void dump_identity (struct hd_driveid *id)
 			printf("off");
 	}
 	printf("\n DblWordIO=%s", YN(id->dword_io&1));
-	dmpstr (", maxPIO=",id->tPIO,SlowMedFast,3);
+	printf(", OldPIO=%d", id->tPIO);
 	printf(", DMA=%s", YN(id->capability&1));
 	if (id->capability&1)
-		dmpstr (", maxDMA=",id->tDMA,SlowMedFast,2);
+		printf(", OldDMA=%d", id->tDMA);
 	putchar('\n');
 	if (!(id->field_valid&1))
 		printf(" (maybe):");
 	printf(" CurCHS=%d/%d/%d, CurSects=%d",
 		id->cur_cyls, id->cur_heads, id->cur_sectors, *(int *)&id->cur_capacity0);
 	printf(", LBA=%s", YN(id->capability&2));
-	if (id->capability&2) {
-		if ((id->cur_sectors == 63) && (id->cur_cyls >= 1024)) {
-			unsigned short new_cyls  = id->cur_cyls;
-			unsigned short new_heads = id->cur_heads;
-
-			if (id->cur_heads == 15) new_heads++;
-
-			do {
-				new_cyls /= 2;
-				new_heads *= 2;
-			} while (new_cyls >=1024);
-
-			if (id->cur_heads == 15)
-				new_heads--;
-
-			printf("\n LBA CHS=%d/%d/%d Remapping",
-				((*(int *)&id->cur_capacity0)/(new_heads * id->cur_sectors)), new_heads, id->cur_sectors);
-		}
-		printf(", LBA=%s", YN(id->capability&2));
+	if (id->capability&2)
  		printf(", LBAsects=%d", id->lba_capacity);
-	}
 	if (id->capability&1) {
 		if ((id->field_valid&2) || (id->dma_1word | id->dma_mword))
 			printf("\n ");
@@ -206,7 +190,17 @@ static void dump_identity (struct hd_driveid *id)
 				if (id->dma_ultra & 0x200)	printf("*");
 				if (id->dma_ultra & 0x002)	printf("mode1 ");
 				if (id->dma_ultra & 0x400)	printf("*");
-				if (id->dma_ultra & 0x004)	printf("mode2");
+				if (id->dma_ultra & 0x004)	printf("mode2 ");
+				if (id->dma_ultra & 0x800)	printf("*");
+				if (id->dma_ultra & 0x008)	printf("mode3 ");
+				if (id->dma_ultra & 0x1000)	printf("*");
+				if (id->dma_ultra & 0x0010)	printf("mode4 ");
+				if (id->dma_ultra & 0x2000)	printf("*");
+				if (id->dma_ultra & 0x0020)	printf("mode5 ");
+				if (id->dma_ultra & 0x4000)	printf("*");
+				if (id->dma_ultra & 0x0040)	printf("mode6 ");
+				if (id->dma_ultra & 0x8000)	printf("*");
+				if (id->dma_ultra & 0x0080)	printf("mode7 ");
 			}
 #endif
 		}
@@ -220,7 +214,7 @@ void flush_buffer_cache (int fd)
 	if (ioctl(fd, BLKFLSBUF, NULL))		/* do it again, big time */
 		perror("BLKFLSBUF failed");
 #ifdef HDIO_DRIVE_CMD
-	if (is_scsi_hd) {
+	if (is_scsi_hd || is_xt_hd) {
 		sleep(1);
 	} else {
 		if (ioctl(fd, HDIO_DRIVE_CMD, NULL))	/* await completion */
@@ -322,9 +316,14 @@ void time_cache (int fd)
 	correction -= (e1.it_value.tv_sec - e2.it_value.tv_sec)
 	 + ((e1.it_value.tv_usec - e2.it_value.tv_usec) / 1000000.0);
 
-	printf("%2d MB in %5.2f seconds =%5.2f MB/sec\n",
-		(BUFCACHE_FACTOR * TIMING_MB), correction,
-		(BUFCACHE_FACTOR * TIMING_MB) / correction);
+	if ((BUFCACHE_FACTOR * timing_MB) >= correction)  /* more than 1MB/s */
+		printf("%2d MB in %5.2f seconds =%6.2f MB/sec\n",
+			(BUFCACHE_FACTOR * timing_MB), correction,
+			(BUFCACHE_FACTOR * timing_MB) / correction);
+	else
+		printf("%2d MB in %5.2f seconds =%6.2f kB/sec\n",
+			(BUFCACHE_FACTOR * timing_MB), correction,
+			(BUFCACHE_FACTOR * timing_MB) / correction * 1024);
 	correction /= BUFCACHE_FACTOR;
 
 	flush_buffer_cache(fd);
@@ -381,8 +380,12 @@ void time_device (int fd)
 
 	elapsed = (e1.it_value.tv_sec - e2.it_value.tv_sec)
 	 + ((e1.it_value.tv_usec - e2.it_value.tv_usec) / 1000000.0);
-	printf("%2d MB in %5.2f seconds =%5.2f MB/sec\n",
-		TIMING_MB, elapsed, TIMING_MB / elapsed);
+	if (timing_MB >= elapsed)  /* more than 1MB/s */
+		printf("%2d MB in %5.2f seconds =%6.2f MB/sec\n",
+			timing_MB, elapsed, timing_MB / elapsed);
+	else
+		printf("%2d MB in %5.2f seconds =%6.2f kB/sec\n",
+			timing_MB, elapsed, timing_MB / elapsed * 1024);
 
 	if (elapsed <= (correction * 2))
 		printf("Hmm.. suspicious results: probably not enough free memory for a proper test.\n");
@@ -390,8 +393,12 @@ void time_device (int fd)
 	else if (correction != 0.0) {
 		printf(" Estimating raw driver speed: ");
 		elapsed -= correction;
-		printf("%2d MB in %5.2f seconds =%5.2f MB/sec\n",
-			TIMING_MB, elapsed, TIMING_MB / elapsed);
+		if (timing_MB >= elapsed)  /* more than 1MB/s */
+			printf("%2d MB in %5.2f seconds =%6.2f MB/sec\n",
+				timing_MB, elapsed, timing_MB / elapsed);
+		else
+			printf("%2d MB in %5.2f seconds =%6.2f kB/sec\n",
+				timing_MB, elapsed, timing_MB / elapsed * 1024);
 	}
 #endif
 quit:
@@ -407,6 +414,14 @@ void no_scsi (void)
 	}
 }
 
+void no_xt (void)
+{
+	if (is_xt_hd) {
+		fputs (" operation not supported on XT disks\n", stderr);
+		exit(EINVAL);
+	}
+}
+
 static void on_off (unsigned int value)
 {
 	printf(value ? " (on)\n" : " (off)\n");
@@ -532,6 +547,12 @@ void process_dev (char *devname)
 	) 
 		is_scsi_hd = 1;
 	else
+#ifdef XT_DISK_MAJOR
+	if (major(stat_buf.st_rdev) == XT_DISK_MAJOR) {
+		is_xt_hd = 1;
+		timing_MB = 2;
+	} else
+#endif
 		if (major(stat_buf.st_rdev) != IDE0_MAJOR
 		 && major(stat_buf.st_rdev) != IDE1_MAJOR
 #ifdef IDE2_MAJOR
@@ -539,9 +560,15 @@ void process_dev (char *devname)
 #endif
 #ifdef IDE3_MAJOR
 		 && major(stat_buf.st_rdev) != IDE3_MAJOR
+#endif
+#ifdef IDE4_MAJOR
+		 && major(stat_buf.st_rdev) != IDE4_MAJOR
+#endif
+#ifdef IDE5_MAJOR
+		 && major(stat_buf.st_rdev) != IDE5_MAJOR
 #endif
 	                                                      ) {
-		fprintf(stderr,"%s is not a hard disk.\n",devname);
+		fprintf(stderr,"%s not supported by hdparm\n",devname);
 		exit(EINVAL);
 	}
 
@@ -550,7 +577,6 @@ void process_dev (char *devname)
 		perror(devname);
 		exit(errno);
 	}
-
 	if (!quiet)
 		printf("\n%s:\n", devname);
 	if (set_readahead) {
@@ -562,6 +588,7 @@ void process_dev (char *devname)
 #ifdef HDIO_SET_PIO_MODE
 	if (set_piomode) {
 		no_scsi();
+		no_xt();
 		if (noisy_piomode) {
 			if (piomode == 255)
 				printf(" attempting to auto-tune PIO mode\n");
@@ -575,6 +602,7 @@ void process_dev (char *devname)
 #ifdef HDIO_SET_32BIT
 	if (set_io32bit) {
 		no_scsi();
+		no_xt();
 		if (get_io32bit)
 			printf(" setting 32-bit I/O support flag to %ld\n", io32bit);
 		if (ioctl(fd, HDIO_SET_32BIT, io32bit)) 
@@ -583,6 +611,7 @@ void process_dev (char *devname)
 #endif
 	if (set_mult) {
 		no_scsi();
+		no_xt();
 		if (get_mult)
 			printf(" setting multcount to %ld\n", mult);
 		if (ioctl(fd, HDIO_SET_MULTCOUNT, mult)) 
@@ -601,6 +630,7 @@ void process_dev (char *devname)
 	}
 	if (set_unmask) {
 		no_scsi();
+		no_xt();
 		if (get_unmask) {
 			printf(" setting unmaskirq to %ld", unmask);
 			on_off(unmask);
@@ -622,6 +652,7 @@ void process_dev (char *devname)
 #ifdef HDIO_SET_NOWERR
 	if (set_nowerr) {
 		no_scsi();
+		no_xt();
 		if (get_nowerr) {
 			printf(" setting nowerr to %ld", nowerr);
 			on_off(nowerr);
@@ -633,6 +664,7 @@ void process_dev (char *devname)
 #ifdef HDIO_GET_KEEPSETTINGS
 	if (set_keep) {
 		no_scsi();
+		no_xt();
 		if (get_keep) {
 			printf(" setting keep_settings to %ld", keep);
 			on_off(keep);
@@ -642,10 +674,23 @@ void process_dev (char *devname)
 	}
 #endif /* HDIO_GET_KEEPSETTINGS */
 #ifdef HDIO_DRIVE_CMD
+	if (set_doorlock) {
+		unsigned char args[4] = {0,0,0,0};
+		no_scsi();
+		no_xt();
+		args[0] = doorlock ? WIN_DOORLOCK : WIN_DOORUNLOCK;
+		if (get_doorlock) {
+			printf(" setting drive doorlock to %ld", doorlock);
+			on_off(doorlock);
+		}
+		if (ioctl(fd, HDIO_DRIVE_CMD, &args))
+			perror(" HDIO_DRIVE_CMD failed");
+	}
 	if (set_dkeep) {
 		/* lock/unlock the drive's "feature" settings */
 		unsigned char args[4] = {WIN_SETFEATURES,0,0,0};
 		no_scsi();
+		no_xt();
 		if (get_dkeep) {
 			printf(" setting drive keep features to %ld", dkeep);
 			on_off(dkeep);
@@ -666,6 +711,7 @@ void process_dev (char *devname)
 	if (set_prefetch) {
 		unsigned char args[4] = {WIN_SETFEATURES,0,0xab,0};
 		no_scsi();
+		no_xt();
 		args[1] = prefetch;
 		if (get_prefetch)
 			printf(" setting drive prefetch to %ld\n", prefetch);
@@ -675,6 +721,7 @@ void process_dev (char *devname)
 	if (set_xfermode) {
 		unsigned char args[4] = {WIN_SETFEATURES,0,3,0};
 		no_scsi();
+		no_xt();
 		args[1] = xfermode;
 		if (get_xfermode) {
 			printf(" setting xfermode to %ld", xfermode);
@@ -686,6 +733,7 @@ void process_dev (char *devname)
 	if (set_lookahead) {
 		unsigned char args[4] = {WIN_SETFEATURES,0,0,0};
 		no_scsi();
+		no_xt();
 		args[2] = lookahead ? 0xaa : 0x55;
 		if (get_lookahead) {
 			printf(" setting drive read-lookahead to %ld", lookahead);
@@ -697,6 +745,7 @@ void process_dev (char *devname)
 	if (set_wcache) {
 		unsigned char args[4] = {WIN_SETFEATURES,0,0,0};
 		no_scsi();
+		no_xt();
 		args[2] = wcache ? 0x02 : 0x82;
 		if (get_wcache) {
 			printf(" setting drive write-caching to %ld", wcache);
@@ -732,6 +781,7 @@ void process_dev (char *devname)
 	if (set_seagate) {
 		unsigned char args[4] = {0xfb,0,0,0};
 		no_scsi();
+		no_xt();
 		if (get_seagate)
 			printf(" disabling Seagate auto powersaving mode\n");
 		if (ioctl(fd, HDIO_DRIVE_CMD, &args))
@@ -740,6 +790,7 @@ void process_dev (char *devname)
 	if (set_standby) {
 		unsigned char args[4] = {WIN_SETIDLE1,standby,0,0};
 		no_scsi();
+		no_xt();
 		if (get_standby) {
 			printf(" setting standby to %ld", standby);
 			interpret_standby(standby);
@@ -759,7 +810,7 @@ void process_dev (char *devname)
 	if (!flagcount)
 		verbose = 1;
 
-	if ((verbose && !is_scsi_hd) || get_mult || get_identity) {
+	if ((verbose && !is_scsi_hd && !is_xt_hd) || get_mult || get_identity) {
 		no_scsi();
 		multcount = -1;
 		if (ioctl(fd, HDIO_GET_MULTCOUNT, &multcount))
@@ -770,8 +821,9 @@ void process_dev (char *devname)
 		}
 	}
 #ifdef HDIO_GET_32BIT
-	if ((verbose && !is_scsi_hd) || get_io32bit) {
+	if ((verbose && !is_scsi_hd && !is_xt_hd) || get_io32bit) {
 		no_scsi();
+		no_xt();
 		if (ioctl(fd, HDIO_GET_32BIT, &parm))
 			perror(" HDIO_GET_32BIT failed");
 		else {
@@ -789,8 +841,9 @@ void process_dev (char *devname)
 		}
 	}
 #endif
-	if ((verbose && !is_scsi_hd) || get_unmask) {
+	if ((verbose && !is_scsi_hd && !is_xt_hd) || get_unmask) {
 		no_scsi();
+		no_xt();
 		if (ioctl(fd, HDIO_GET_UNMASKINTR, &parm))
 			perror(" HDIO_GET_UNMASKINTR failed");
 		else {
@@ -810,8 +863,9 @@ void process_dev (char *devname)
 	}
 #endif
 #ifdef HDIO_GET_KEEPSETTINGS
-	if ((verbose && !is_scsi_hd) || get_keep) {
+	if ((verbose && !is_scsi_hd && !is_xt_hd) || get_keep) {
 		no_scsi();
+		no_xt();
 		if (ioctl(fd, HDIO_GET_KEEPSETTINGS, &parm))
 			perror(" HDIO_GET_KEEPSETTINGS failed");
 		else {
@@ -822,8 +876,9 @@ void process_dev (char *devname)
 #endif /* HDIO_GET_KEEPSETTINGS */
 
 #ifdef HDIO_SET_NOWERR
-	if (verbose || get_nowerr) {
+	if ((verbose && !is_scsi_hd && !is_xt_hd) || get_nowerr) {
 		no_scsi();
+		no_xt();
 		if (ioctl(fd, HDIO_GET_NOWERR, &parm))
 			perror(" HDIO_GET_NOWERR failed");
 		else {
@@ -879,6 +934,7 @@ void process_dev (char *devname)
 #endif
 	if (get_identity) {
 		no_scsi();
+		no_xt();
 #ifdef HDIO_OBSOLETE_IDENTITY
 		if (!(old_identity = ioctl(fd, HDIO_GET_IDENTITY, &id)) || !ioctl(fd, HDIO_OBSOLETE_IDENTITY)) {
 #else
@@ -898,16 +954,18 @@ void process_dev (char *devname)
 	}
 #ifdef HDIO_SET_32BIT
 	if (get_IDentity) {
-		unsigned char args[4+512] = {WIN_PIDENTIFY,0,0,1,};
+		unsigned char args[4+512] = {WIN_IDENTIFY,0,0,1,};
 		no_scsi();
+		no_xt();
 		if (!(ioctl(fd, HDIO_DRIVE_CMD, &args))) {
 			old_identity = 0;
 			dump_identity((struct hd_driveid *)&args[4]);
 		} else {
-			args[0] = WIN_IDENTIFY;
-			if (!(ioctl(fd, HDIO_DRIVE_CMD, &args)))
+			args[0] = WIN_PIDENTIFY;
+			if (!(ioctl(fd, HDIO_DRIVE_CMD, &args))) {
+				old_identity = 0;
 				dump_identity((struct hd_driveid *)&args[4]);
-			else
+			} else
 				perror(" HDIO_DRIVE_CMD failed");
 		}
 	}
@@ -923,7 +981,7 @@ void process_dev (char *devname)
 
 void usage_error (void)
 {
-	fprintf(stderr,"\n%s - get/set hard disk parameters - version 3.5\n\n", progname);
+	fprintf(stderr,"\n%s - get/set hard disk parameters - version %s\n\n", progname, VERSION);
 	fprintf(stderr,"Usage:  %s  [options] [device] ..\n\n", progname);
 	fprintf(stderr,"Options:\n"
 	" -a    get/set fs readahead\n"
@@ -952,6 +1010,7 @@ void usage_error (void)
 #endif
 #ifdef HDIO_DRIVE_CMD
 	" -K  * set drive keep_features_over_reset flag (0/1)\n"
+	" -L  * set drive doorlock (0/1) (removable harddisks only)\n"
 #endif
 	" -m  * get/set multiple sector count\n"
 #ifdef HDIO_SET_NOWERR 
@@ -971,7 +1030,8 @@ void usage_error (void)
 	" -t    perform device read timings\n"
 	" -T    perform cache read timings\n"
 	" -u  * get/set unmaskirq flag (0/1)\n"
-	" -v    default; same as -acdgkmnru (-gr for SCSI)\n"
+	" -v    default; same as -acdgkmnru (-gr for SCSI, -adgr for XT)\n"
+	" -V    display program version and exit immediately\n"
 #ifdef HDIO_DRIVE_CMD
 	" -W  * set drive write-caching flag (0/1) (DANGEROUS)\n"
 	" -X  * set IDE xfer mode (DANGEROUS)\n"
@@ -1011,6 +1071,10 @@ int main(int argc, char **argv)
 			while ((c = *p++)) {
 				++flagcount;
 				switch (c) {
+					case 'V':
+						fprintf(stdout, "%s %s\n", progname, VERSION);
+						exit(0);
+						break;
 					case 'v':
 						verbose = 1;
 						break;
@@ -1158,6 +1222,18 @@ int main(int argc, char **argv)
 							fprintf(stderr, "-A: missing value (0/1)\n");
 						break;
 
+					case 'L':
+						get_doorlock = noisy;
+						noisy = 1;
+						if (!*p && argc && isdigit(**argv))
+							p = *argv++, --argc;
+						if (*p == '0' || *p == '1') {
+							set_doorlock = 1;
+							doorlock = *p++ - '0';
+						} else
+							fprintf(stderr, "-L: missing value (0/1)\n");
+						break;
+
 					case 'W':
 						get_wcache = noisy;
 						noisy = 1;
